@@ -495,6 +495,8 @@ class SMPLSequence(Node):
 
     @selected_mode.setter
     def selected_mode(self, selected_mode):
+        if self._selected_mode == selected_mode:
+            return
         self._selected_mode = selected_mode
 
         if self.selected_mode == 'edit':
@@ -605,16 +607,22 @@ class SMPLSequence(Node):
                 self.selected_mode = 'edit'
                 imgui.close_current_popup()
 
-    def on_selection(self, node, tri_id):
+            imgui.spacing()
+            imgui.separator()
+            imgui.spacing()
+            super().gui_context_menu(imgui)
+
+    def on_selection(self, node, instance_id, tri_id):
         if self.edit_mode:
-            # Find the index of the joint that is currently being edited.
-            self._edit_joint = self.rbs.get_index_from_node_and_triangle(node, tri_id)
-            if self._edit_joint is not None:
+            # Index of the joint that is currently being edited.
+            if node != self.mesh_seq:
+                self._edit_joint = instance_id
                 self.rbs.color_one(self._edit_joint, (0.3, 0.4, 1, 1))
             else:
+                self._edit_joint = None
                 # Reset color of all spheres to the default color
                 self.rbs.color = self.rbs.color
 
-    def render_outline(self, ctx, camera, prog):
+    def render_outline(self, *args, **kwargs):
         # Only render outline of the mesh, skipping skeleton and rigid bodies.
-        self.mesh_seq.render_outline(ctx, camera, prog)
+        self.mesh_seq.render_outline(*args, **kwargs)
