@@ -189,6 +189,14 @@ class Camera(Node, CameraInterface):
             if self.path[1] is not None:
                 self.path[1].enabled = enabled
 
+    @property
+    def bounds(self):
+        return self.mesh.bounds
+
+    @property
+    def current_bounds(self):
+        return self.mesh.current_bounds
+
     def hide_frustum(self):
         if self.frustum:
             self.remove(self.frustum)
@@ -297,10 +305,10 @@ class Camera(Node, CameraInterface):
         self.path = (path_spheres, path_lines)
         self.current_frame_id = frame_id
 
-    def render_outline(self, ctx, camera, prog):
+    def render_outline(self, *args, **kwargs):
         # Only render the mesh outline, this avoids outlining
         # the frustum and coordinate system visualization.
-        self.mesh.render_outline(ctx, camera, prog)
+        self.mesh.render_outline(*args, **kwargs)
 
     def view_from_camera(self):
         """If the viewer is specified for this camera, change the current view to view from this camera"""
@@ -362,6 +370,7 @@ class WeakPerspectiveCamera(Camera):
 
         assert scale.shape[0] == translation.shape[0], "Number of frames in scale and translation must match"
 
+        kwargs['gui_affine'] = False
         super(WeakPerspectiveCamera, self).__init__(n_frames=scale.shape[0], viewer=viewer, **kwargs)
 
         self.scale_factor = scale
@@ -434,6 +443,11 @@ class WeakPerspectiveCamera(Camera):
             else:
                 self.hide_frustum()
 
+        imgui.spacing()
+        imgui.separator()
+        imgui.spacing()
+        super(Camera, self).gui_context_menu(imgui)
+
 
 class OpenCVCamera(Camera):
     """ A camera described by extrinsics and intrinsics in the format used by OpenCV """
@@ -459,6 +473,7 @@ class OpenCVCamera(Camera):
         assert self.K.shape[0] == 1 or self.Rt.shape[0] == 1 or self.K.shape[0] == self.Rt.shape[0], (
             f"extrinsics and intrinsics array shape mismatch: {self.Rt.shape} and {self.K.shape}")
 
+        kwargs['gui_affine'] = False
         super(OpenCVCamera, self).__init__(viewer=viewer, n_frames=max(self.K.shape[0], self.Rt.shape[0]), **kwargs)
         self.position = self.current_position
         self.rotation = self.current_rotation
@@ -612,6 +627,11 @@ class OpenCVCamera(Camera):
             else:
                 self.hide_frustum()
 
+        imgui.spacing()
+        imgui.separator()
+        imgui.spacing()
+        super(Camera, self).gui_context_menu(imgui)
+
 
 class PinholeCamera(Camera):
     """
@@ -658,6 +678,7 @@ class PinholeCamera(Camera):
     def rotation(self):
         return np.array([-self.right, self.up, -self.forward]).T
 
+
     def update_matrices(self, width, height):
         # Compute projection matrix.
         P = perspective_projection(np.deg2rad(self.fov), width / height, self.near, self.far)
@@ -698,6 +719,13 @@ class PinholeCamera(Camera):
 
         return OpenCVCamera(K, Rts, cols, rows, near=self.near, far=self.far, viewer=self.viewer, **kwargs)
 
+    def gui_affine(self, imgui):
+        """ Render GUI for affine transformations"""
+        # Position controls
+        u, pos = imgui.drag_float3('Position##pos{}'.format(self.unique_name), *self.position, 0.1, format='%.2f')
+        if u:
+            self.position = pos
+
     @hooked
     def gui(self, imgui):
         u, show = imgui.checkbox("Show frustum", self.frustum is not None)
@@ -716,6 +744,11 @@ class PinholeCamera(Camera):
             else:
                 self.hide_frustum()
 
+        imgui.spacing()
+        imgui.separator()
+        imgui.spacing()
+        super(Camera, self).gui_context_menu(imgui)
+
 
 class ViewerCamera(CameraInterface):
     """
